@@ -9,31 +9,31 @@
 #define SRC_CLUSTERING_INFOMAPBASE_H_
 
 #include "InfomapConfig.h"
-#include <string>
-#include <iostream>
-#include <sstream>
 #include "InfoEdge.h"
+#include "InfoNode.h"
+#include "InfomapIterator.h"
+#include "ClusterMap.h"
+#include "../io/Network.h"
+#include "../utils/Log.h"
+#include "../utils/Date.h"
+#include "../utils/Stopwatch.h"
+
 #include <vector>
 #include <deque>
 #include <map>
-#include "../utils/Log.h"
 #include <limits>
-#include "PartitionQueue.h"
-#include <limits>
-// #include "StateNetwork.h"
-#include "../io/Network.h"
-#include "InfoNode.h"
-#include "InfomapIterator.h"
-#include "./ClusterMap.h"
-#include "../utils/Date.h"
-#include "../utils/Stopwatch.h"
+#include <string>
+#include <iostream>
+#include <sstream>
 
 namespace infomap {
 
 struct PerLevelStat;
-// class InfoNode;
 
-//template<typename Node, template<typename,typename> class Optimizer = GreedyOptimizer<Node, MapEquation>>
+namespace detail {
+  class PartitionQueue;
+} // namespace detail
+
 class InfomapBase : public InfomapConfig<InfomapBase> {
   template <typename Objective>
   friend class InfomapOptimizer;
@@ -42,12 +42,11 @@ protected:
   using EdgeType = Edge<InfoNode>;
 
 public:
-  InfomapBase() : InfomapConfig<InfomapBase>() {}
+  using PartitionQueue = detail::PartitionQueue;
 
-  explicit InfomapBase(const Config& conf) : InfomapConfig<InfomapBase>(conf),
-                                             m_network(conf)
-  {
-  }
+  InfomapBase() : InfomapConfig<InfomapBase>() { }
+
+  explicit InfomapBase(const Config& conf) : InfomapConfig<InfomapBase>(conf), m_network(conf) { }
 
   explicit InfomapBase(const std::string flags) : InfomapConfig<InfomapBase>(flags)
   {
@@ -175,11 +174,6 @@ public:
   bool isSuperLevelOnTopLevel() const;
   bool isMainInfomap() const;
 
-  virtual bool haveMemory() const
-  {
-    return isMemoryNetwork();
-  }
-
   bool haveHardPartition() const;
 
   std::vector<InfoNode*>& activeNetwork() const;
@@ -425,6 +419,15 @@ public:
    */
   std::string writeFlowTree(std::string filename = "", bool states = false);
 
+	/**
+	 * Write Newick tree to a .tre file.
+	 * @param filename the filename for the output file. If empty, use default 
+	 * based on output directory and input file name
+	 * @param states if memory network, print the state-level network without merging physical nodes within modules
+	 * @return the filename written to
+	 */
+	std::string writeNewickTree(std::string filename = "", bool states = false);
+
   /**
    * Write tree to a .clu file.
    * @param filename the filename for the output file. If empty, use default
@@ -437,18 +440,6 @@ public:
    */
   std::string writeClu(std::string filename = "", bool states = false, int moduleIndexLevel = 1);
 
-  /**
-   * Write modular network to a .map file.
-   * @param filename the filename for the output file. If empty, use default
-   * based on output directory and input file name
-   * @param states if memory network, print the state-level network without merging physical nodes within modules
-   * @param moduleIndexLevel the depth from the root on which to advance module index.
-   * Value 1 will give the module index on the coarsest level, 2 the level below and so on. Default
-   * value -1 will give the module index for the lowest level, i.e. the finest modular structure.
-   * @return the filename written to
-   */
-  std::string writeMap(std::string filename = "", bool states = false, int moduleIndexLevel = -1);
-
   /**
     * Print per level statistics
     * @param out The output stream to print the per level statistics to
@@ -483,6 +474,12 @@ protected:
    */
   void writeTreeLinks(std::ostream& outStream, bool states = false);
 
+	/**
+	 * Write Newick tree to output stream
+	 * @param states, write state-level tree, else aggregate physical nodes within modules
+	 */
+	void writeNewickTree(std::ostream& outStream, bool states = false);
+
   InfoNode m_root;
   std::vector<InfoNode*> m_leafNodes;
   std::vector<InfoNode*> m_moduleNodes;
@@ -517,42 +514,58 @@ protected:
 
 
 struct PerLevelStat {
-  PerLevelStat() = default;
   double codelength() const { return indexLength + leafLength; }
+
   unsigned int numNodes() const { return numModules + numLeafNodes; }
+
   unsigned int numModules = 0;
   unsigned int numLeafNodes = 0;
   double indexLength = 0.0;
   double leafLength = 0.0;
 };
 
-struct PerIterationStats {
-  PerIterationStats()
-      : iterationIndex(0),
-        numTopModules(0),
-        numBottomModules(0),
-        topPerplexity(0.0),
-        bottomPerplexity(0.0),
-        topOverlap(0.0),
-        bottomOverlap(0.0),
-        codelength(0.0),
-        maxDepth(0),
-        weightedDepth(0.0),
-        seconds(0.0),
-        isMinimum(false) {}
-  unsigned int iterationIndex;
-  unsigned int numTopModules;
-  unsigned int numBottomModules;
-  double topPerplexity; // Perplexity of top module flow distribution
-  double bottomPerplexity;
-  double topOverlap; // Average number of modules per physical node
-  double bottomOverlap;
-  double codelength;
-  unsigned int maxDepth;
-  double weightedDepth;
-  double seconds;
-  bool isMinimum;
-};
+
+namespace detail {
+
+  class PartitionQueue {
+    using PendingModule = InfoNode*;
+
+  public:
+    using size_t = std::deque<PendingModule>::size_type;
+
+    void swap(PartitionQueue& other)
+    {
+      std::swap(level, other.level);
+      std::swap(numNonTrivialModules, other.numNonTrivialModules);
+      std::swap(flow, other.flow);
+      std::swap(nonTrivialFlow, other.nonTrivialFlow);
+      std::swap(skip, other.skip);
+      std::swap(indexCodelength, other.indexCodelength);
+      std::swap(leafCodelength, other.leafCodelength);
+      std::swap(moduleCodelength, other.moduleCodelength);
+      m_queue.swap(other.m_queue);
+    }
+
+    size_t size() const { return m_queue.size(); }
+
+    void resize(size_t size) { m_queue.resize(size); }
+
+    PendingModule& operator[](size_t i) { return m_queue[i]; }
+
+    unsigned int level = 1;
+    unsigned int numNonTrivialModules = 0;
+    double flow = 0.0;
+    double nonTrivialFlow = 0.0;
+    bool skip = false;
+    double indexCodelength = 0.0; // Consolidated
+    double leafCodelength = 0.0; // Consolidated
+    double moduleCodelength = 0.0; // Left to improve on next level
+
+  private:
+    std::deque<PendingModule> m_queue;
+  };
+
+} // namespace detail
 
 } // namespace infomap
 
