@@ -6,42 +6,39 @@
  */
 
 #include "InfomapBase.h"
-#include <string>
-#include <iostream>
-#include <sstream>
+#include "InfomapConfig.h"
 #include "InfoNode.h"
 #include "InfoEdge.h"
-#include <vector>
-#include <deque>
 #include "FlowData.h"
 #include "MapEquation.h"
+#include "ClusterMap.h"
 #include "../io/SafeFile.h"
+#include "../io/version.h"
+#include "../io/Network.h"
 #include "../utils/Log.h"
 #include "../utils/infomath.h"
 #include "../utils/Date.h"
 #include "../utils/Stopwatch.h"
 #include "../utils/exceptions.h"
 #include "../utils/FileURI.h"
-#include "../io/version.h"
+
+#include <string>
+#include <vector>
 #include <iomanip>
 #include <limits>
-#include "InfomapConfig.h"
-#include "ClusterMap.h"
 #include <map>
 #include <set>
-// #include <type_traits>
-#include "PartitionQueue.h"
-#include <cstdlib> // abs
-#include <limits>
-#include <algorithm> // std::sort
-// #include "StateNetwork.h"
-#include "../io/Network.h"
+#include <cstdlib>
+#include <algorithm>
+
 #ifdef _OPENMP
 #include <omp.h>
 #endif
 
 namespace infomap {
 
+using detail::PartitionQueue;
+
 // ===================================================
 // IO
 // ===================================================
@@ -112,6 +109,7 @@ bool InfomapBase::haveNonTrivialModules() const
 
 unsigned int InfomapBase::numLevels() const
 {
+  // TODO: Make sure this is not called unless tree is guaranteed to have even depth!
   unsigned int depth = 0;
   InfoNode* n = m_root.firstChild;
   while (n != nullptr) {
@@ -223,7 +221,7 @@ void InfomapBase::run(std::string parameters)
   std::string currentParameters = io::Str() << m_initialParameters << (parameters.empty() ? "" : " ") << parameters;
   if (currentParameters != m_currentParameters) {
     m_currentParameters = currentParameters;
-    setConfig(Config::fromString(m_currentParameters, isCLI));
+    setConfig(Config(m_currentParameters, isCLI));
     m_network.setConfig(*this);
   }
 
@@ -303,14 +301,14 @@ void InfomapBase::run(Network& network)
 
   if (network.haveMemoryInput()) {
     Log() << "  -> Found higher order network input, using the Map Equation for higher order network flows\n";
-    if (!isMemoryNetwork()) {
+    if (!haveMemory()) {
       setMemoryInput();
     }
     if (network.isMultilayerNetwork() && !isMultilayerNetwork()) {
       setMultilayerInput();
     }
   } else {
-    if (isMemoryNetwork()) {
+    if (haveMemory()) {
       Log() << "  -> Warning: Higher order network specified but no higher order input found.\n";
     }
     Log() << "  -> Ordinary network input, using the Map Equation for first order network flows\n";
@@ -394,7 +392,7 @@ void InfomapBase::run(Network& network)
         if (numTrials > 1) {
           bestTree.clear();
           for (auto it(iterLeafNodes()); !it.isEnd(); ++it) {
-            bestTree.add(it->stateId, it.path());
+            bestTree.emplace_back(it->stateId, it.path());
           }
         }
       }
@@ -488,15 +486,13 @@ InfomapBase& InfomapBase::initPartition(std::string clusterDataFile, bool hard)
 
   Log() << "Init partition from file '" << clusterDataFile << "'... ";
 
-  if (clusterMap.extension() == "tree") {
+  const auto& ext = clusterMap.extension();
+
+  if (ext == "tree" || ext == "ftree") {
     initTree(clusterMap.nodePaths());
-    if (!noInfomap) {
-      noInfomap = true;
-      //TODO: implement continuing from tree
-      Log() << "\nNotice: Continuing clustering algorithm after reading input tree currently not supported, assuming --no-infomap for now.\n";
-    }
-  } else if (clusterMap.extension() == "clu")
+  } else if (ext == "clu") {
     initPartition(clusterMap.clusterIds(), hard);
+  }
 
   Log() << "done! Generated " << numLevels() << " levels with codelength " << m_hierarchicalCodelength << "\n";
 
@@ -506,9 +502,9 @@ InfomapBase& InfomapBase::initPartition(std::string clusterDataFile, bool hard)
 InfomapBase& InfomapBase::initTree(const NodePaths& tree)
 {
   Log(4) << "Init tree... " << std::setprecision(9);
-  unsigned int maxDepth = 2;
+  auto maxDepth = 2;
   std::map<unsigned int, unsigned int> nodeIdToIndex;
-  unsigned int leafIndex = 0;
+  auto leafIndex = 0;
   for (auto& leafNode : m_leafNodes) {
     // Also detach leaf nodes to delete all modules, safe to call multiple times
     leafNode->parent->releaseChildren();
@@ -519,15 +515,16 @@ InfomapBase& InfomapBase::initTree(const NodePaths& tree)
   }
   m_root.deleteChildren();
 
-  unsigned int numNodesFound = 0;
-  unsigned int numNodesNotInNetwork = 0;
-  for (auto& nodePath : tree.nodePaths) {
+  auto numNodesFound = 0;
+  auto numNodesNotInNetwork = 0;
+  for (const auto& nodePath : tree) {
     ++numNodesFound;
     InfoNode* node = &root();
-    unsigned int depth = 0;
-    auto& path = nodePath.path;
-    auto nodeId = nodePath.nodeId;
+    auto depth = 0;
+    const auto& path = nodePath.second;
+    const auto nodeId = nodePath.first;
     InfoNode* leafNode = nullptr;
+
     try {
       auto nodeIndex = nodeIdToIndex.at(nodeId);
       leafNode = m_leafNodes[nodeIndex];
@@ -535,8 +532,10 @@ InfomapBase& InfomapBase::initTree(const NodePaths& tree)
       ++numNodesNotInNetwork;
       continue;
     }
-    for (unsigned int i = 0; i < path.size(); ++i) {
+
+    for (size_t i = 0; i < path.size(); ++i) {
       auto childNumber = path[i]; // 1-based indexing
+
       // Create new node if path doesn't exist
       // TODO: Check correct tree indexing?
       if (node->childDegree() < childNumber) {
@@ -546,21 +545,26 @@ InfomapBase& InfomapBase::initTree(const NodePaths& tree)
         }
         node->addChild(child);
       }
+
       node = node->lastChild;
       ++depth;
     }
     maxDepth = std::max(maxDepth, depth);
   }
+
   if (numNodesNotInNetwork > 0) {
     Log(1) << "\n -> " << numNodesNotInNetwork << "/" << numNodesFound << " nodes in tree not found in network.";
   }
-  unsigned int numNodesAddedToNeighbouringModules = 0;
-  unsigned int numNodesWithoutClusterInfo = 0;
+
+  auto numNodesAddedToNeighbouringModules = 0;
+  auto numNodesWithoutClusterInfo = 0;
+
   // Set orphaned nodes to their own or neighbouring module
   for (auto& leafNode : m_leafNodes) {
     if (leafNode->parent != nullptr) {
       continue;
     }
+
     ++numNodesWithoutClusterInfo;
     if (assignToNeighbouringModule) {
       // Take first neighbour that has a module assigned
@@ -969,6 +973,90 @@ void InfomapBase::hierarchicalPartition()
 {
   Log(1) << "Hierarchical partition..." << std::endl;
 
+  const auto depth = maxTreeDepth();
+  if (depth > 2) {
+    Log(1) << "Continuing from a tree with " << depth << " levels..." << std::endl;
+
+    if (fastHierarchicalSolution == 0) {
+      Log(1) << "Removing sub modules...\n";
+      removeSubModules(true);
+      m_hierarchicalCodelength = calcCodelengthOnTree(true);
+    }
+
+    else if (fastHierarchicalSolution == 1) {
+      Log(1) << "Fine-tune bottom modules... ";
+      bool isSilent = isMainInfomap() && Log::isSilent();
+
+      double codelengthBefore = 0.0;
+      double codelengthAfter = 0.0;
+
+      for (InfoNode& module : m_root.infomapTree()) {
+        if (!module.isLeaf() && module.firstChild->isLeafModule()) {
+          codelengthBefore += module.codelength;
+          auto numLeafs = 0;
+          for (auto& leafModule : module) {
+            numLeafs += leafModule.childDegree();
+          }
+
+          std::vector<unsigned int> modules(numLeafs);
+          std::vector<InfoNode*> leafs(numLeafs);
+
+          auto i = 0;
+          for (auto it = module.begin_tree(); !it.isEnd(); ++it) {
+            if (it->isLeaf()) {
+              modules[i] = it.moduleIndex();
+              leafs[i] = it.current();
+              ++i;
+            }
+          }
+          
+          module.replaceChildrenWithGrandChildren();
+          
+          auto& subInfomap = getSubInfomap(module).initNetwork(module);
+          
+          subInfomap.initPartition(modules);
+
+          // Run two-level partition + find hierarchically super modules (skip recursion)
+          subInfomap.setOnlySuperModules(true).run();
+          
+          // Collect sub Infomap modules
+          i = 0;
+          for (auto& subLeafPtr : subInfomap.leafNodes()) {
+            modules[i] = subLeafPtr->index;
+            ++i;
+          }
+
+          // Create new sub modules
+          std::vector<InfoNode*> subModules(numLeafs, nullptr);
+          module.releaseChildren();
+
+          for (auto i = 0; i < numLeafs; ++i) {
+            InfoNode* leaf = leafs[i];
+            unsigned int moduleIndex = modules[i];
+            if (subModules[moduleIndex] == nullptr) {
+              subModules[moduleIndex] = new InfoNode(subInfomap.leafNodes()[i]->parent->data);
+              subModules[moduleIndex]->index = moduleIndex;
+              module.addChild(subModules[moduleIndex]);
+            }
+            subModules[moduleIndex]->addChild(leaf);
+          }
+          module.disposeInfomap();
+          module.codelength = calcCodelength(module);
+          codelengthAfter += module.codelength;
+        }
+      }
+
+      if (isMainInfomap())
+        Log::setSilent(isSilent);
+
+      const double diffCodelength = codelengthBefore - codelengthAfter;
+      Log() << "done! Codelength improvement " << (diffCodelength / codelengthBefore) * 100 << "% to codelength " << codelengthAfter << "\n";
+    }
+
+    recursivePartition();
+    return;
+  }
+
   partition();
 
   if (numTopModules() == 1 || numTopModules() == numLeafNodes()) {
@@ -978,20 +1066,16 @@ void InfomapBase::hierarchicalPartition()
 
   if (numTopModules() > preferredNumberOfModules) {
     findHierarchicalSuperModules();
-    //	findHierarchicalSuperModulesFast(superLevelLimit);
   }
 
-  // printRSS();
-
   if (onlySuperModules) {
     removeSubModules(true);
-    m_hierarchicalCodelength = calcCodelengthOnTree(true); // FIX? TODO: What needs to be fixed?
+    m_hierarchicalCodelength = calcCodelengthOnTree(true);
     return;
   }
 
   if (fastHierarchicalSolution >= 2) {
-    // Calculate individual module codelengths and store on the modules
-    // calcCodelengthOnTree(false);
+    // FIXME Calculate individual module codelengths and store on the modules?
     return;
   }
 
@@ -1001,7 +1085,6 @@ void InfomapBase::hierarchicalPartition()
   }
 
   recursivePartition();
-  // printRSS();
 }
 
 void InfomapBase::partition()
@@ -1392,6 +1475,7 @@ unsigned int InfomapBase::coarseTune()
       InfomapBase& subInfomap = getSubInfomap(node)
                                     .setTwoLevel(true)
                                     .setTuneIterationLimit(1);
+      // subInfomap.preferModularSolution = true;//TODO: Benchmark more
       subInfomap.initNetwork(node).run();
 
       auto originalLeafIt = node.begin_child();
@@ -1957,6 +2041,16 @@ void InfomapBase::writeResult()
   if (noFileOutput)
     return;
 
+  // Log() << "\n\nNetwork flow:\n";
+  // for (auto& leafNode : m_leafNodes) {
+  //   auto& n = *leafNode;
+  //   Log() << n.stateId << " (" << n.data.flow << "):\n";
+  //   for (auto& edge : n.outEdges()) {
+  //     auto& target = edge->target;
+  //     Log() << " -> " << target.stateId << " (" << edge->data.flow << ")\n";
+  //   }
+  // }
+
   // Log() << "\nPhysical tree:\n";
   // for (InfomapIteratorPhysical it(&root()); !it.isEnd(); ++it) {
   // 	Log() << io::stringify(it.path(), ":") << " moduleIndex: " << it.moduleIndex() << ", stateId: " << it->stateId <<
@@ -2007,6 +2101,26 @@ void InfomapBase::writeResult()
   }
 
 
+  if (printNewick) {
+    std::string filename = outDirectory + outName + ".tre";
+
+    if (!haveMemory()) {
+      Log() << "Write Newick tree to " << filename << "... ";
+      writeNewickTree(filename);
+      Log() << "done!\n";
+    } else {
+      // Write both physical and state level
+      Log() << "Write physical Newick tree to " << filename << "... ";
+      writeNewickTree(filename, false);
+      Log() << "done!\n";
+      std::string filenameStates = outDirectory + outName + "_states.tre";
+      Log() << "Write state Newick tree to " << filenameStates << "... ";
+      writeNewickTree(filenameStates, true);
+      Log() << "done!\n";
+    }
+  }
+
+
   if (printClu) {
     std::string filename = outDirectory + outName + ".clu";
     if (!haveMemory()) {
@@ -2024,35 +2138,18 @@ void InfomapBase::writeResult()
       Log() << "done!\n";
     }
   }
-
-  if (printMap) {
-    std::string filename = outDirectory + outName + ".map";
-    if (!haveMemory()) {
-      Log() << "Write modular network to " << filename << "... ";
-      writeMap(filename);
-      Log() << "done!\n";
-    } else {
-      // Write both physical and state level
-      Log() << "Write physical modular network to " << filename << "... ";
-      writeMap(filename);
-      Log() << "done!\n";
-      std::string filenameStates = outDirectory + outName + "_states.map";
-      Log() << "Write state modular network to " << filenameStates << "... ";
-      writeMap(filenameStates, true);
-      Log() << "done!\n";
-    }
-  }
 }
 
 std::string InfomapBase::getOutputFileHeader()
 {
+  std::string bipartiteInfo = io::Str() << "\n# bipartite start id " << m_network.bipartiteStartId();
   return io::Str() << "# v" << INFOMAP_VERSION << "\n"
                    << "# ./Infomap " << parsedString << "\n"
                    << "# started at " << m_startDate << "\n"
                    << "# completed in " << m_elapsedTime.getElapsedTimeInSec() << " s\n"
                    << "# partitioned into " << maxTreeDepth() << " levels with " << numTopModules() << " top modules\n"
                    << "# codelength " << codelength() << " bits\n"
-                   << "# relative codelength savings " << getRelativeCodelengthSavings() * 100 << "%";
+                   << "# relative codelength savings " << getRelativeCodelengthSavings() * 100 << "%" << (m_network.isBipartite() ? bipartiteInfo : "");
 }
 
 std::string InfomapBase::writeTree(std::string filename, bool states)
@@ -2076,28 +2173,47 @@ std::string InfomapBase::writeFlowTree(std::string filename, bool states)
   return outputFilename;
 }
 
+std::string InfomapBase::writeNewickTree(std::string filename, bool states)
+{
+  std::string outputFilename = filename.empty() ? outDirectory + outName + (haveMemory() && states ? "_states.tre" : ".tre") : filename;
+
+  SafeOutFile outFile(outputFilename);
+  writeNewickTree(outFile, states);
+
+  return outputFilename;
+}
+
 std::string InfomapBase::writeClu(std::string filename, bool states, int moduleIndexLevel)
 {
   // unsigned int maxModuleLevel = maxTreeDepth();
   std::string outputFilename = filename.empty() ? outDirectory + outName + (haveMemory() && states ? "_states.clu" : ".clu") : filename;
   SafeOutFile outFile(outputFilename);
+
   outFile << std::setprecision(9);
   outFile << getOutputFileHeader() << "\n";
-  outFile << "# module level: " << moduleIndexLevel << "\n";
+  outFile << "# module level " << moduleIndexLevel << "\n";
   outFile << std::resetiosflags(std::ios::floatfield) << std::setprecision(6);
+
   if (states) {
     outFile << "# state_id module flow node_id";
     if (isMultilayerNetwork())
       outFile << " layer_id";
-    outFile << "\n";
+    outFile << '\n';
   } else {
     outFile << "# node_id module flow\n";
   }
-  // auto it = haveMemory() && !states ? iterTreePhysical(moduleIndexLevel) : iterTree(moduleIndexLevel);
+
+  const auto shouldHideBipartiteNodes = isBipartite() && hideBipartiteNodes;
+  const auto bipartiteStartId = shouldHideBipartiteNodes ? m_network.bipartiteStartId() : 0;
+
   if (haveMemory() && !states) {
     for (auto it(iterTreePhysical(moduleIndexLevel)); !it.isEnd(); ++it) {
       InfoNode& node = *it;
       if (node.isLeaf()) {
+        if (shouldHideBipartiteNodes && node.physicalId >= bipartiteStartId) {
+          continue;
+        }
+
         outFile << node.physicalId << " " << it.moduleId() << " " << node.data.flow << "\n";
       }
     }
@@ -2105,6 +2221,10 @@ std::string InfomapBase::writeClu(std::string filename, bool states, int moduleI
     for (auto it(iterTree(moduleIndexLevel)); !it.isEnd(); ++it) {
       InfoNode& node = *it;
       if (node.isLeaf()) {
+        if (shouldHideBipartiteNodes && node.physicalId >= bipartiteStartId) {
+          continue;
+        }
+
         if (states) {
           outFile << node.stateId << " " << it.moduleId() << " " << node.data.flow << " " << node.physicalId;
           if (isMultilayerNetwork())
@@ -2118,158 +2238,73 @@ std::string InfomapBase::writeClu(std::string filename, bool states, int moduleI
   return outputFilename;
 }
 
-std::string InfomapBase::writeMap(std::string filename, bool states, int moduleIndexLevel)
-{
-  //TODO: Support moduleIndexLevel other than default -1
-  if (moduleIndexLevel != -1) {
-    Log() << "\nWarning: writeMap currently only supports default moduleIndexLevel -1\n";
-  }
-  // unsigned int maxModuleLevel = maxTreeDepth();
-  std::string outputFilename = filename.empty() ? outDirectory + outName + (haveMemory() && states ? "_states.map" : ".map") : filename;
-
-  // Collect modules
-  std::vector<InfoNode*> modules;
-  for (auto it(iterLeafModules()); !it.isEnd(); ++it) {
-    modules.push_back(it.current());
-  }
-
-  // Sort modules on flow descending order
-  std::sort(modules.begin(), modules.end(), [](InfoNode* a, InfoNode* b) {
-    return a->data.flow > b->data.flow;
-  });
-
-  // Store id on modules to simplify link aggregation
-  for (unsigned int i = 0; i < modules.size(); ++i) {
-    modules[i]->index = i + 1;
-  }
-
-  // Copy node data to seperate structure as physical nodes may be ephemeral
-  struct MapNode {
-    unsigned int id;
-    double flow;
-    std::string name = "";
-    MapNode(unsigned int id, double flow, std::string name = "")
-        : id(id), flow(flow), name(name) { initName(); }
-    void initName()
-    {
-      if (name.empty()) {
-        name = io::stringify(id);
-      }
-    }
-  };
-  // Collect nodes
-  unsigned int numNodes = 0;
-  std::vector<std::vector<MapNode>> nodes(modules.size());
-  for (unsigned int i = 0; i < modules.size(); ++i) {
-    if (haveMemory() && !states) {
-      for (InfomapLeafIteratorPhysical it(modules[i]); !it.isEnd(); ++it) {
-        // Use physicalId as node id for (aggregated) physical nodes
-        nodes[i].push_back(MapNode(it->physicalId, it->data.flow, m_network.names()[it->physicalId]));
-      }
-    } else {
-      // State nodes
-      for (auto& node : *modules[i]) {
-        nodes[i].push_back(MapNode(node.stateId, node.data.flow, m_network.names()[node.stateId]));
-      }
-    }
-    // Sort nodes on flow descending order
-    std::sort(nodes[i].begin(), nodes[i].end(), [=](MapNode& a, MapNode& b) {
-      return a.flow > b.flow;
-    });
-    numNodes += nodes[i].size();
-  }
-
-  // Collect links and module exit flow
-  std::map<std::pair<unsigned int, unsigned int>, double> moduleLinks;
-  std::map<unsigned int, double> moduleExitFlow;
-  for (auto& module : modules) {
-    for (auto& node : *module) {
-      for (auto& link : node.outEdges()) {
-        if (link->source.parent != link->target.parent) {
-          moduleLinks[std::make_pair(link->source.parent->index, link->target.parent->index)] += link->data.flow;
-          moduleExitFlow[module->index + 1] += link->data.flow;
-        }
-      }
-    }
-  }
-
-  SafeOutFile outFile(outputFilename);
-  outFile << "# modules: " << modules.size() << "\n";
-  outFile << "# modulelinks: " << moduleLinks.size() << "\n";
-  outFile << "# nodes: " << numNodes << "\n";
-  outFile << "# links: " << 0 << "\n";
-  outFile << std::setprecision(9);
-  outFile << "# codelength: " << m_hierarchicalCodelength << "\n";
-  outFile << std::setprecision(6);
-  outFile << (isUndirectedClustering() ? "*Undirected\n" : "*Directed\n");
-  outFile << "*Modules " << modules.size() << "\n";
-  for (unsigned int i = 0; i < modules.size(); ++i) {
-    auto& module = *modules[i];
-    //# id name flow exitFlow (Name the module from the biggest child)
-    outFile << module.index << " \"" << nodes[i][0].name << ",...\" " << module.data.flow << " " << moduleExitFlow[module.index] << "\n";
-  }
-  outFile << "*Nodes " << numNodes << "\n";
-  for (unsigned int i = 0; i < nodes.size(); ++i) {
-    auto& siblings = nodes[i];
-    for (unsigned int j = 0; j < siblings.size(); ++j) {
-      auto& n = siblings[j];
-      outFile << i + 1 << ":" << j + 1 << " \"" << n.name << "\" " << n.flow << "\n";
-      // outFile << " " << n.id << "\n";
-    }
-  }
-  outFile << "*Links " << moduleLinks.size() << "\n";
-  for (auto& linkIt : moduleLinks) {
-    auto& link = linkIt.first;
-    auto& weight = linkIt.second;
-    outFile << link.first << " " << link.second << " " << weight << "\n";
-  }
-  return outputFilename;
-}
-
 void InfomapBase::writeTree(std::ostream& outStream, bool states)
 {
   auto oldPrecision = outStream.precision();
   outStream << std::setprecision(9);
   outStream << getOutputFileHeader() << "\n";
   outStream << std::setprecision(6);
+
   if (states) {
     outStream << "# path flow name state_id node_id";
     if (isMultilayerNetwork())
       outStream << " layer_id";
-    outStream << "\n";
-  } else
+    outStream << '\n';
+  } else {
     outStream << "# path flow name node_id\n";
+  }
+
+  const auto nodeName = [this](const auto& node) {
+    auto name = m_network.names()[node.physicalId];
+
+    if (name.empty()) {
+      name = io::stringify(node.physicalId);
+    }
+
+    return name;
+  };
+
+  const auto shouldHideBipartiteNodes = !printFlowTree && isBipartite() && hideBipartiteNodes;
+  const auto bipartiteStartId = shouldHideBipartiteNodes ? m_network.bipartiteStartId() : 0;
+
   // TODO: Make a general iterator where merging physical nodes depend on a parameter rather than type to be able to DRY here
   if (haveMemory() && !states) {
     for (auto it(iterTreePhysical()); !it.isEnd(); ++it) {
       InfoNode& node = *it;
       if (node.isLeaf()) {
+        if (shouldHideBipartiteNodes && node.physicalId >= bipartiteStartId) {
+          continue;
+        }
+
         auto& path = it.path();
-        auto name = m_network.names()[node.physicalId];
-        if (name.empty())
-          name = io::stringify(node.physicalId);
-        outStream << io::stringify(path, ":") << " " << node.data.flow << " \"" << name << "\" " << node.physicalId << "\n";
+
+        outStream << io::stringify(path, ":") << " " << node.data.flow << " \"" << nodeName(node) << "\" " << node.physicalId << '\n';
       }
     }
   } else {
     for (auto it(iterTree()); !it.isEnd(); ++it) {
       InfoNode& node = *it;
       if (node.isLeaf()) {
+        if (shouldHideBipartiteNodes && node.physicalId >= bipartiteStartId) {
+          continue;
+        }
+
         auto& path = it.path();
-        auto name = m_network.names()[node.physicalId];
-        if (name.empty())
-          name = io::stringify(node.physicalId);
-        outStream << io::stringify(path, ":") << " " << node.data.flow << " \"" << name << "\" ";
+
+        outStream << io::stringify(path, ":") << " " << node.data.flow << " \"" << nodeName(node) << "\" ";
+
         if (states) {
           outStream << node.stateId << " " << node.physicalId;
           if (isMultilayerNetwork())
             outStream << " " << node.layerId;
-          outStream << "\n";
-        } else
-          outStream << node.physicalId << "\n";
+          outStream << '\n';
+        } else {
+          outStream << node.physicalId << '\n';
+        }
       }
     }
   }
+
   outStream << std::setprecision(oldPrecision);
 }
 
@@ -2286,63 +2321,71 @@ void InfomapBase::writeTreeLinks(std::ostream& outStream, bool states)
   // Map state id to parent in infomap tree iterator
   std::map<unsigned int, InfoNode*> stateIdToParent;
   std::map<unsigned int, unsigned int> stateIdToChildIndex;
+
   if (mergePhysicalNodes) {
     for (auto it(iterTreePhysical()); !it.isEnd(); ++it) {
-      if (it->isLeaf()) {
-        for (auto stateId : it->stateNodes) {
-          stateIdToParent[stateId] = it->parent;
-          stateIdToChildIndex[stateId] = it.childIndex();
+      auto& node = *it;
+      if (node.isLeaf()) {
+        for (auto stateId : node.stateNodes) {
+          stateIdToParent[stateId] = node.parent;
+          stateIdToChildIndex[stateId] = node.childIndex();
         }
       } else {
         // Use stateId to store depth on modules to simplify link aggregation
-        it->stateId = it.depth();
-        it->index = it.childIndex();
+        node.stateId = node.depth();
+        node.index = node.childIndex();
       }
     }
   } else {
     for (auto it(iterTree()); !it.isEnd(); ++it) {
-      if (it->isLeaf()) {
-        stateIdToParent[it->stateId] = it->parent;
-        stateIdToChildIndex[it->stateId] = it.childIndex();
+      auto& node = *it;
+
+      if (node.isLeaf()) {
+        stateIdToParent[node.stateId] = node.parent;
+        stateIdToChildIndex[node.stateId] = node.childIndex();
       } else {
         // Use stateId to store depth on modules to simplify link aggregation
-        it->stateId = it.depth();
-        it->index = it.childIndex();
+        node.stateId = node.depth();
+        node.index = node.childIndex();
       }
     }
   }
 
-  using LinkMap = std::map<std::pair<unsigned int, unsigned int>, double>;
+  using Link = std::pair<unsigned int, unsigned int>;
+  using LinkMap = std::map<Link, double>;
   std::map<std::string, LinkMap> moduleLinks;
 
   for (auto& leaf : m_leafNodes) {
     for (auto& link : leaf->outEdges()) {
-      // Log() << link->source.stateId << " " << link->target.stateId << " " << link->data.flow << "\n";
       double flow = link->data.flow;
-      // auto& sourceIt = stateIdToParent[link->source.stateId];
-      // auto& targetIt = stateIdToParent[link->target.stateId];
       InfoNode* sourceParent = stateIdToParent[link->source.stateId];
       InfoNode* targetParent = stateIdToParent[link->target.stateId];
+
       auto sourceDepth = sourceParent->calculatePath().size() + 1;
       auto targetDepth = targetParent->calculatePath().size() + 1;
+
       auto sourceChildIndex = stateIdToChildIndex[link->source.stateId];
       auto targetChildIndex = stateIdToChildIndex[link->target.stateId];
+
       auto sourceParentIt = InfomapParentIterator(sourceParent);
       auto targetParentIt = InfomapParentIterator(targetParent);
-      // Log() << io::stringify(sourceIt.path(), ":", 1) << " (" << link->source.stateId << ") -> " << io::stringify(targetIt.path(), ":", 1) << " (" << link->target.stateId << "), flow: " << flow << "\n";
+
       // Iterate to same depth
       // First raise target
       while (targetDepth > sourceDepth) {
         ++targetParentIt;
         --targetDepth;
       }
+
       // Raise source to same depth
       while (sourceDepth > targetDepth) {
         ++sourceParentIt;
         --sourceDepth;
       }
+
       auto currentDepth = sourceDepth;
       // Add link if same parent
+
       while (currentDepth > 0) {
         if (sourceParentIt == targetParentIt) {
           // Skip self-links
@@ -2352,10 +2395,13 @@ void InfomapBase::writeTreeLinks(std::ostream& outStream, bool states)
             linkMap[std::make_pair(sourceChildIndex + 1, targetChildIndex + 1)] += flow;
           }
         }
+
         sourceChildIndex = sourceParentIt->index;
         targetChildIndex = targetParentIt->index;
+
         ++sourceParentIt;
         ++targetParentIt;
+
         --currentDepth;
       }
     }
@@ -2369,11 +2415,9 @@ void InfomapBase::writeTreeLinks(std::ostream& outStream, bool states)
     auto parentId = io::stringify(it.path(), ":");
     auto& module = *it;
     auto& links = moduleLinks[parentId];
-    // if (it->isLeafModule() && mergePhysicalNodes) {
-    // 	outStream << "*Links " << parentId << " " << exitFlow[parentId] << " " << 0 << " " << 0 << "\n";
-    // 	continue;
-    // }
-    outStream << "*Links " << (parentId == "" ? "root" : parentId) << " " << module.data.enterFlow << " " << module.data.exitFlow << " " << links.size() << " " << module.infomapChildDegree() << "\n";
+
+    outStream << "*Links " << (parentId.empty() ? "root" : parentId) << " " << module.data.enterFlow << " " << module.data.exitFlow << " " << links.size() << " " << module.infomapChildDegree() << "\n";
+
     for (auto itLink : links) {
       unsigned int sourceId = itLink.first.first;
       unsigned int targetId = itLink.first.second;
@@ -2381,10 +2425,65 @@ void InfomapBase::writeTreeLinks(std::ostream& outStream, bool states)
       outStream << sourceId << " " << targetId << " " << flow << "\n";
     }
   }
+
   outStream << std::setprecision(oldPrecision);
 }
 
 
+void InfomapBase::writeNewickTree(std::ostream& outStream, bool states)
+{
+  auto oldPrecision = outStream.precision();
+  // outStream << std::setprecision(9);
+  // outStream << getOutputFileHeader() << "\n";
+  outStream << std::setprecision(6);
+
+  auto isRoot = true;
+  unsigned int lastDepth = 0;
+  std::vector<double> flowStack;
+
+  auto writeNewickNode = [&](std::ostream& o, const InfoNode& node, unsigned int depth) {
+    if (depth > lastDepth || isRoot) {
+      outStream << "(";
+      flowStack.push_back(node.data.flow);
+      if (node.isLeaf())
+        outStream << (states ? node.stateId : node.physicalId) << ":" << node.data.flow;
+    } else if (depth == lastDepth) {
+      outStream << ",";
+      flowStack[flowStack.size()-1] = node.data.flow;
+      if (node.isLeaf()) {
+        outStream << (states ? node.stateId : node.physicalId) << ":" << node.data.flow;
+      }
+    } else {
+      // depth < lastDepth
+      while (flowStack.size() > depth + 1) {
+        flowStack.pop_back();
+        outStream << "):" << flowStack.back();
+      }
+      flowStack[flowStack.size()-1] = node.data.flow;
+      outStream << ",";
+    }
+    lastDepth = depth;
+    isRoot = false;
+  };
+  
+  // TODO: Make a general iterator where merging physical nodes depend on a parameter rather than type to be able to DRY here
+  if (haveMemory() && !states) {
+    for (auto it(iterTreePhysical()); !it.isEnd(); ++it) {
+      writeNewickNode(outStream, *it, it.depth());
+    }
+  } else {
+    for (auto it(iterTree()); !it.isEnd(); ++it) {
+      writeNewickNode(outStream, *it, it.depth());
+    }
+  }
+  while (flowStack.size() > 1) {
+    flowStack.pop_back();
+    outStream << "):" << flowStack.back();
+  }
+  outStream << ");\n";
+  outStream << std::setprecision(oldPrecision);
+}
+
 unsigned int InfomapBase::printPerLevelCodelength(std::ostream& out)
 {
   std::vector<PerLevelStat> perLevelStats;
