@@ -590,6 +590,11 @@ void Network::printSummary()
 
 void Network::addMultilayerLink(unsigned int layer1, unsigned int n1, unsigned int layer2, unsigned int n2, double weight)
 {
+  if (weight < m_config.weightThreshold) {
+    ++m_numLinksIgnoredByWeightThreshold;
+    m_totalLinkWeightIgnored += weight;
+    return;
+  }
   unsigned int stateId1 = addMultilayerNode(layer1, n1);
   unsigned int stateId2 = addMultilayerNode(layer2, n2);
 
@@ -732,8 +737,7 @@ void Network::generateStateNetworkFromMultilayerWithInterLinks()
 
 void Network::generateStateNetworkFromMultilayerWithSimulatedInterLinks()
 {
-  Log() << "Generating state network from multilayer networks with simulated inter-layer links...\n"
-        << std::flush;
+  Log() << "Generating state network from multilayer networks with simulated inter-layer links...\n" << std::flush;
   double relaxRate = m_config.multilayerRelaxRate;
 
   int maxRelaxLimit = m_networks.size();
@@ -756,6 +760,125 @@ void Network::generateStateNetworkFromMultilayerWithSimulatedInterLinks()
     return layer1 >= layer2 ? diff <= relaxLimitDown : -diff <= relaxLimitUp;
   };
 
+  if (m_config.multilayerRelaxByJensenShannonDivergence) {
+    Log() << "-> Using Jensen-Shannon Divergence\n";
+
+    for (unsigned int nodeId = 0; nodeId <= m_maxNodeIdInIntraLayerNetworks; ++nodeId)
+    {
+
+      unsigned int layer2from = 0;
+      
+      // Calculate Jensen-Shannon similarity between all layers such that layer1 >= layer2,
+      // and then use its symmetry for layer2 > layer1
+      std::map<unsigned int,std::map<unsigned int,double> > jsRelaxWeights;
+      std::map<unsigned int,double> jsTotWeight;
+
+      for (unsigned int layer1 = 0; layer1 < m_networks.size(); ++layer1)
+      {
+
+        unsigned int layer2to = layer1+1;
+        // Limit possible jumps to close by layers
+        if(m_config.multilayerRelaxLimit >= 0) {
+          layer2from = ((int)layer1-m_config.multilayerRelaxLimit) < 0 ? 0 : layer1-m_config.multilayerRelaxLimit;
+        }
+
+        auto& layer1LinkMap = m_networks[layer1].nodeLinkMap();
+        auto& layer1OutLinks = layer1LinkMap[StateNode(nodeId)];
+        // Skip dangling nodes, because they have no information to calculate similarity
+        if (layer1OutLinks.empty())
+          continue;
+        
+        double sumOutLinkWeightLayer1 = m_networks[layer1].outWeights()[nodeId];
+
+
+        for (unsigned int layer2 = layer2from; layer2 < layer2to; ++layer2){
+          auto& layer2LinkMap = m_networks[layer2].nodeLinkMap();
+          auto& layer2OutLinks = layer2LinkMap[StateNode(nodeId)];
+          if (layer2OutLinks.empty())
+            continue;
+
+          double sumOutLinkWeightLayer2 = m_networks[layer2].outWeights()[nodeId];
+  
+          bool intersect;
+          double div = calculateJensenShannonDivergence(intersect,layer1OutLinks,sumOutLinkWeightLayer1,layer2OutLinks,sumOutLinkWeightLayer2);
+          double jsWeight = 1.0 - div;
+          if(intersect && (jsWeight >= m_config.multilayerJSRelaxLimit)){
+            jsTotWeight[layer1] += jsWeight;
+            jsRelaxWeights[layer1][layer2] = jsWeight;
+            if(layer1 != layer2){
+              jsTotWeight[layer2] += jsWeight;
+              jsRelaxWeights[layer2][layer1] = jsWeight;
+            }
+          }
+        }
+      }
+
+      // Second loop over all pairs of layers
+      unsigned int layer2to = m_networks.size();
+
+      for (unsigned int layer1 = 0; layer1 < m_networks.size(); ++layer1)
+      {
+        // Limit possible jumps to close by layers
+        if(m_config.multilayerRelaxLimit >= 0){
+          layer2from = ((int)layer1-m_config.multilayerRelaxLimit) < 0 ? 0 : layer1-m_config.multilayerRelaxLimit;
+          layer2to = (layer1+m_config.multilayerRelaxLimit) > m_networks.size() ? m_networks.size() : layer1+m_config.multilayerRelaxLimit;
+        }
+
+        double sumOutLinkWeightLayer1 = m_networks[layer1].outWeights()[nodeId];
+
+        std::map<unsigned int,std::map<unsigned int,double> >::iterator jsRelaxWeightsLayer1It = jsRelaxWeights.find(layer1);
+        std::map<unsigned int,double>::iterator jsTotWeightIt = jsTotWeight.find(layer1);
+
+        // Create inter-links to the intra-connected nodes in other layers
+        for (unsigned int layer2 = layer2from; layer2 < layer2to; ++layer2)
+        {
+          if(jsRelaxWeightsLayer1It != jsRelaxWeights.end()){
+            std::map<unsigned int,double>::iterator jsRelaxWeightsIt = jsRelaxWeightsLayer1It->second.find(layer2);
+            if(jsRelaxWeightsIt != jsRelaxWeightsLayer1It->second.end()){
+
+              bool isIntra = layer2 == layer1;
+      
+              // Create inter-links to the outgoing nodes in the target layer
+              double linkWeightNormalizationFactor;
+              if (isIntra){
+                linkWeightNormalizationFactor = 1;
+              } else {
+                linkWeightNormalizationFactor = jsRelaxWeightsIt->second * relaxRate / (1.0 - relaxRate) * sumOutLinkWeightLayer1 / jsTotWeightIt->second;
+              }
+              
+              // double stateNodeWeightNormalizationFactor = 1.0;
+              // createIntraLinksToNeighbouringNodesInTargetLayer(layer1, nodeId, layer2, m_networks[layer2].nodeLinkMap(), linkWeightNormalizationFactor, stateNodeWeightNormalizationFactor);
+
+              auto& targetLinks = m_networks[layer2].nodeLinkMap();
+              auto& targetOutlinks = targetLinks[StateNode(nodeId)];
+              if (targetOutlinks.empty()) {
+                // Log() << "   -> Dangling\n";
+                continue;
+              }
+              for (auto& outLink : targetOutlinks) {
+                auto& n2 = outLink.first.physicalId;
+                auto& linkData = outLink.second;
+                double intraWeight = linkData.weight;
+                // Add intra link weight as teleport weight to source node
+                //asdf
+                unsigned int stateId1 = addMultilayerNode(layer1, nodeId, intraWeight);
+                unsigned int stateId2i = addMultilayerNode(layer2, n2, 0.0);
+
+                // Log() << "      -> " << n2 << "\n";
+                double weight = intraWeight == 0.0 ? 0.0 : linkWeightNormalizationFactor * intraWeight;
+                addLink(stateId1, stateId2i, weight);
+                ++m_numInterLayerLinks; 
+              }
+            }
+          }
+        }
+      }
+    }
+
+    return;
+  }
+
+
   for (auto& it1 : m_networks) {
     auto layer1 = it1.first;
     auto& network1 = it1.second;
@@ -820,6 +943,86 @@ void Network::generateStateNetworkFromMultilayerWithSimulatedInterLinks()
   }
 }
 
+double Network::calculateJensenShannonDivergence(bool &intersect, const OutLinkMap& layer1OutLinks, double sumOutLinkWeightLayer1, const OutLinkMap& layer2OutLinks, double sumOutLinkWeightLayer2)
+{
+	intersect = false;
+	double h1 = 0.0; // The entropy rate of the node in the first layer
+	double h2 = 0.0; // The entropy rate of the node in the second layer
+	double h12 = 0.0; // The entropy rate of the lumped node
+	// The out-link weights of the nodes
+	double ow1 = sumOutLinkWeightLayer1;
+	double ow2 = sumOutLinkWeightLayer2;
+	// Normalized weights over node in layer 1 and 2
+	double pi1 = ow1 / (ow1 + ow2);
+	double pi2 = ow2 / (ow1 + ow2);
+
+	auto layer1OutLinkIt = layer1OutLinks.begin();
+	auto layer2OutLinkIt = layer2OutLinks.begin();
+	auto layer1OutLinkItEnd = layer1OutLinks.end();
+	auto layer2OutLinkItEnd = layer2OutLinks.end();
+	while(layer1OutLinkIt != layer1OutLinkItEnd && layer2OutLinkIt != layer2OutLinkItEnd){
+		auto diff = layer1OutLinkIt->first.id - layer2OutLinkIt->first.id;		
+		if(diff < 0){
+		// If the first state node has a link that the second has not
+			double p1 = layer1OutLinkIt->second.weight/ow1;
+			h1 -= p1*log2(p1);
+			double p12 = pi1*layer1OutLinkIt->second.weight/ow1;
+			h12 -= p12*log2(p12);
+			layer1OutLinkIt++;
+		}
+		else if(diff > 0){
+		// If the second state node has a link that the second has not
+			double p2 = layer2OutLinkIt->second.weight/ow2;
+			h2 -= p2*log2(p2);
+			double p12 = pi2*layer2OutLinkIt->second.weight/ow2;
+			h12 -= p12*log2(p12);
+			layer2OutLinkIt++;
+		}
+		else{ // If both state nodes have the link
+			intersect = true;
+			double p1 = layer1OutLinkIt->second.weight/ow1;
+			h1 -= p1*log2(p1);
+			double p2 = layer2OutLinkIt->second.weight/ow2;
+			h2 -= p2*log2(p2);
+			double p12 = pi1*layer1OutLinkIt->second.weight/ow1 + pi2*layer2OutLinkIt->second.weight/ow2;
+			h12 -= p12*log2(p12);
+			layer1OutLinkIt++;
+			layer2OutLinkIt++;
+		}
+	}
+
+	while(layer1OutLinkIt != layer1OutLinkItEnd){
+		// If the first state node has a link that the second has not
+		double p1 = layer1OutLinkIt->second.weight/ow1;
+		h1 -= p1*log2(p1);
+		double p12 = pi1*layer1OutLinkIt->second.weight/ow1;
+		h12 -= p12*log2(p12);
+		layer1OutLinkIt++;
+	}
+
+	while(layer2OutLinkIt != layer2OutLinkItEnd){
+		// If the second state node has a link that the second has not
+		double p2 = layer2OutLinkIt->second.weight/ow2;
+		h2 -= p2*log2(p2);
+		double p12 = pi2*layer2OutLinkIt->second.weight/ow2;
+		h12 -= p12*log2(p12);
+		layer2OutLinkIt++;
+	}	
+	
+	double div = (pi1+pi2)*h12 - pi1*h1 - pi2*h2;
+
+	// Fix precision problems
+	if(div < 0.0)
+		div = 0.0;
+	else if(div > 1.0)
+		div = 1.0;
+	
+	// Log() << "\n" << div << " " << (pi1+pi2) << " " << h12 << " " << pi1 << " " << h1 << " " << pi2 << " " << h2;
+
+	return div;
+
+}
+
 void Network::simulateInterLayerLinks()
 {
 }
@@ -830,6 +1033,7 @@ void Network::addMultilayerIntraLink(unsigned int layer, unsigned int n1, unsign
   bool added = m_networks[layer].addLink(n1, n2, weight);
   if (added) {
     ++m_numIntraLayerLinks;
+    m_maxNodeIdInIntraLayerNetworks = std::max(m_maxNodeIdInIntraLayerNetworks, std::max(n1, n2));
   }
 }
 
@@ -870,7 +1074,7 @@ void Network::addMultilayerInterLink(unsigned int layer1, unsigned int n, unsign
   // }
 }
 
-unsigned int Network::addMultilayerNode(unsigned int layerId, unsigned int physicalId)
+unsigned int Network::addMultilayerNode(unsigned int layerId, unsigned int physicalId, double weight)
 {
   // auto layerNode = LayerNode(layerId, physicalId);
   // auto it = m_layerNodeToStateId.find(layerNode);
@@ -893,6 +1097,7 @@ unsigned int Network::addMultilayerNode(unsigned int layerId, unsigned int physi
   auto ret = addStateNodeWithAutogeneratedId(physicalId);
   auto& stateNode = ret.first->second;
   stateNode.layerId = layerId;
+  stateNode.weight = weight;
   m_layerNodeToStateId[layerId][physicalId] = stateNode.id;
   m_layers.insert(layerId);
   return stateNode.id;
